// Copyright (c) 2019, TheDOme6
//
// Please see the included LICENSE file for more information.

import 'dart:async';
import 'dart:convert';
import 'package:http/http.dart' as http;
import 'dart:io';
import 'dart:math';

class TRTLServices {
  final String _host = 'https://api.trtl.services/v1';
  final String _token;
  Duration _timeout;

  TRTLServices(this._token, [this._timeout = const Duration(seconds: 2)]);

  /// Create a new TRTL address.
  ///
  /// Throws a [TimeoutException] if the given [Duration] is reached.
  Future<Map<String, dynamic>> createAddress() async {
    return await _post('address', {});
  }

  /// Get details of an address.
  ///
  /// Throws a [TimeoutException] if the given [Duration] is reached
  /// or Throws a [ArgumentError.notNull] if [address] is null.
  Future<Map<String, dynamic>> getAddress(String address) async {
    _isNull(address, 'Address');
    return await _get('address/$address');
  }

  /// Delete an address.
  ///
  /// or Throws a [ArgumentError.notNull] if [address] is null.
  Future<Map<String, dynamic>> deleteAddress(String address) async {
    _isNull(address, 'Address');
    return await _delete('address/$address');
  }

  /// Get all addresses.
  ///
  /// Throws a [TimeoutException] if the given [Duration] is reached.
  Future<List<dynamic>> getAddresses() async {
    return await _get('address/all');
  }

  /// Scan an address for transactions between a 100 block range.
  ///
  /// Throws a [TimeoutException] if the given [Duration] is reached
  /// or Throws a [ArgumentError.notNull] if [address] or [blockIndex] is null.
  Future<List<dynamic>> scanAddress(String address, int blockIndex) async {
    _isNull(address, 'Address');
    _isNull(blockIndex, 'Block index');
    return await _get('address/scan/$address/$blockIndex');
  }

  /// Get the public and secret spend key of an address.
  ///
  /// Throws a [TimeoutException] if the given [Duration] is reached
  /// or Throws a [ArgumentError.notNull] if [address] is null.
  Future<Map<String, dynamic>> getAddressKeys(String address) async {
    _isNull(address, 'Address');
    return await _get('address/keys/$address');
  }

  /// Create an integrated address with an address and payment ID.
  ///
  /// Throws a [TimeoutException] if the given [Duration] is reached
  /// or Throws a [ArgumentError.notNull] if [address] or [paymentId] is null.
  Future<Map<String, dynamic>> integrateAddress(
      String address, String paymentId) async {
    _isNull(address, 'Address');
    _isNull(paymentId, 'Payment Id');
    return await _post(
        'address/integrate', {'address': address, 'paymentId': paymentId});
  }

  /// Get all integrated addresses belonging to an address.
  ///
  /// Throws a [TimeoutException] if the given [Duration] is reached
  /// or Throws a [ArgumentError.notNull] if [address] is null.
  Future<Map<String, dynamic>> getIntegratedAddresses(String address) async {
    _isNull(address, 'Address');
    return await _get('address/integrate/$address');
  }

  /// Calculate the fee in TRTL for a wallet container.
  ///
  /// Throws a [TimeoutException] if the given [Duration] is reached
  /// or Throws a [ArgumentError.notNull] if [amount] is null.
  Future<num> getFee(num amount) async {
    _isNull(amount, 'Amount');
    return await _get('transfer/fee/$amount');
  }

  /// Create a new transaction.
  ///
  /// Throws a [TimeoutException] if the given [Duration] is reached
  /// or Throws a [ArgumentError.notNull] if [sender], [receiver] or
  /// [amount] is null.
  Future<Map<String, dynamic>> createTransfer(
      String sender, String receiver, num amount,
      [num fee = 0.1, String paymentId, String extra]) async {
    _isNull(sender, 'Sender');
    _isNull(receiver, 'Receiver');
    _isNull(amount, 'Amount');
    _isNull(fee, 'Fee');
    return await _post('transfer', {
      'from': sender,
      'to': receiver,
      'amount': amount,
      'fee': fee,
      'paymentId': paymentId ?? null,
      'extra': extra ?? null
    });
  }

  /// Get a transaction created with TRTL Services by hash.
  ///
  /// Throws a [TimeoutException] if the given [Duration] is reached
  /// or Throws a [ArgumentError.notNull] if [hash] is null.
  Future<Map<String, dynamic>> getTransfer(String hash) async {
    _isNull(hash, 'Hash');
    return await _get('transfer/$hash');
  }

  /// Get wallet container info and health check.
  ///
  /// Throws a [TimeoutException] if the given [Duration] is reached.
  Future<Map<String, dynamic>> getWallet() async {
    return await _get('wallet');
  }

  /// Get TRTL Services infrastructure health check.
  ///
  /// Throws a [TimeoutException] if the given [Duration] is reached.
  Future<List<dynamic>> getStatus() async {
    return await _get('status');
  }

  /// Generate a random Payment Id
  static String generatePaymentId() {
    String letters = 'abcdefghijklmnopqrstuvwxyz1234567890';
    String combine = '';
    for (var i = 0; i < 64; i++) {
      int choose = Random.secure().nextInt(letters.length) + 1;
      combine += letters.substring(choose - 1, choose);
    }
    return combine;
  }

  void _isNull(dynamic proof, String required) {
    proof ?? (throw new ArgumentError.notNull(required));
  }

  Future<dynamic> _delete(String uri) async {
    dynamic result = json.decode((await http.delete('$_host/$uri', headers: {
      HttpHeaders.authorizationHeader: 'Bearer $_token'
    }).timeout(_timeout))
        .body);
    if (result is String) {
      throw Exception('TRTLServices responded with an error: $result');
    }
    return result;
  }

  Future<dynamic> _get(String uri) async {
    dynamic result = json.decode((await http.get('$_host/$uri', headers: {
      HttpHeaders.authorizationHeader: 'Bearer $_token'
    }).timeout(_timeout))
        .body);
    if (result is String) {
      throw Exception('TRTLServices responded with an error: $result');
    }
    return result;
  }

  Future<dynamic> _post(String uri, Map<String, dynamic> options) async {
    dynamic result = json.decode((await http.post('$_host/$uri',
            body: json.encode(options),
            headers: {
          HttpHeaders.authorizationHeader: 'Bearer $_token'
        }).timeout(_timeout))
        .body);
    if (result is String) {
      throw Exception('TRTLServices responded with an error: $result');
    }
    return result;
  }
}
